from unittest.mock import patch

import pytest
from infoblox_client import objects

from gso.products import ProductName, Site
from gso.products.product_blocks.router import RouterRole
from gso.products.product_types.router import Router
from gso.services.subscriptions import get_product_id_by_name
from gso.utils.shared_enums import Vendor
from test import USER_CONFIRM_EMPTY_FORM
from test.services.conftest import MockedSharePointClient
from test.workflows import (
    assert_complete,
    assert_lso_interaction_failure,
    assert_lso_interaction_success,
    assert_suspended,
    extract_state,
    resume_workflow,
    run_workflow,
)


@pytest.fixture()
def router_creation_input_form_data(site_subscription_factory, faker):
    router_site = site_subscription_factory()

    return {
        "tt_number": faker.tt_number(),
        "router_site": router_site,
        "hostname": faker.pystr(),
        "ts_port": faker.pyint(),
        "router_role": faker.random_choices(elements=(RouterRole.P, RouterRole.PE, RouterRole.AMT), length=1)[0],
        "vendor": Vendor.NOKIA,
    }


@pytest.mark.workflow()
@patch("gso.utils.workflow_steps.lso_client.execute_playbook")
@patch("gso.workflows.router.create_router.NetboxClient.create_device")
@patch("gso.workflows.router.create_router.infoblox.hostname_available")
@patch("gso.workflows.router.create_router.infoblox.find_host_by_fqdn")
@patch("gso.workflows.router.create_router.infoblox.allocate_host")
@patch("gso.workflows.router.create_router.SharePointClient")
@patch("gso.workflows.router.create_router.KentikClient")
def test_create_nokia_router_success(
    mock_kentik_client,
    mock_sharepoint_client,
    mock_allocate_host,
    mock_find_host_by_fqdn,
    mock_hostname_available,
    mock_netbox_create_device,
    mock_provision_router,
    router_creation_input_form_data,
    faker,
    data_config_filename,
):
    #  Set up mock return values
    product_id = get_product_id_by_name(ProductName.ROUTER)
    mock_site = Site.from_subscription(router_creation_input_form_data["router_site"]).site
    mock_v4 = faker.ipv4()
    mock_v6 = faker.ipv6()
    mock_fqdn = (
        f"{router_creation_input_form_data["hostname"]}.{mock_site.site_name.lower()}."
        f"{mock_site.site_country_code.lower()}.geant.net"
    )
    mock_hostname_available.return_value = True
    mock_allocate_host.return_value = str(mock_v4), str(mock_v6)
    mock_sharepoint_client.return_value = MockedSharePointClient

    #  Run workflow
    initial_router_data = [{"product": product_id}, router_creation_input_form_data]
    result, process_stat, step_log = run_workflow("create_router", initial_router_data)

    state = extract_state(result)
    subscription_id = state["subscription_id"]
    mock_find_host_by_fqdn.return_value = objects.HostRecord(
        connector=None,
        aliases=[mock_fqdn],
        comment=subscription_id,
        ipv4addrs=[
            objects.IPv4(
                ipv4addr=str(mock_v4),
                configure_for_dhcp=False,
                mac="00:00:00:00:00:00",
                ip=str(mock_v4),
                host=f"lo0.{mock_fqdn}",
            ),
        ],
        name=mock_fqdn,
    )

    for _ in range(2):
        result, step_log = assert_lso_interaction_success(result, process_stat, step_log)

    # Handle four consecutive user input steps
    for _ in range(4):
        assert_suspended(result)
        result, step_log = resume_workflow(process_stat, step_log, input_data=USER_CONFIRM_EMPTY_FORM)

    result, step_log = assert_lso_interaction_success(result, process_stat, step_log)

    assert_suspended(result)
    result, step_log = resume_workflow(process_stat, step_log, input_data=USER_CONFIRM_EMPTY_FORM)

    assert_complete(result)

    state = extract_state(result)
    subscription = Router.from_subscription(subscription_id)

    assert subscription.status == "provisioning"
    assert subscription.description == f"Router {mock_fqdn}"

    assert mock_provision_router.call_count == 3
    assert mock_netbox_create_device.call_count == 1
    assert mock_find_host_by_fqdn.call_count == 1
    assert mock_sharepoint_client.call_count == 1
    assert mock_kentik_client.call_count == 1
    assert "ipam_warning" not in state


@pytest.mark.workflow()
@patch("gso.utils.workflow_steps.lso_client.execute_playbook")
@patch("gso.workflows.router.create_router.NetboxClient.create_device")
@patch("gso.workflows.router.create_router.infoblox.hostname_available")
@patch("gso.workflows.router.create_router.infoblox.find_network_by_cidr")
@patch("gso.workflows.router.create_router.infoblox.find_host_by_fqdn")
@patch("gso.workflows.router.create_router.infoblox.allocate_v6_network")
@patch("gso.workflows.router.create_router.infoblox.allocate_v4_network")
@patch("gso.workflows.router.create_router.infoblox.allocate_host")
def test_create_nokia_router_lso_failure(
    mock_allocate_host,
    mock_allocate_v4_network,
    mock_allocate_v6_network,
    mock_find_host_by_fqdn,
    mock_find_network_by_cidr,
    mock_hostname_available,
    mock_netbox_create_device,
    mock_provision_router,
    router_creation_input_form_data,
    faker,
):
    #  Set up mock return values
    mock_site = Site.from_subscription(router_creation_input_form_data["router_site"]).site
    mock_v4 = faker.ipv4()
    mock_v4_net = faker.ipv4(network=True)
    mock_v6 = faker.ipv6()
    mock_fqdn = (
        f"{router_creation_input_form_data["hostname"]}.{mock_site.site_name.lower()}."
        f"{mock_site.site_country_code.lower()}.geant.net"
    )
    mock_hostname_available.return_value = True
    mock_allocate_host.return_value = str(mock_v4), str(mock_v6)
    mock_allocate_v4_network.return_value = mock_v4_net
    mock_allocate_v6_network.return_value = faker.ipv6(network=True)
    mock_find_host_by_fqdn.return_value = objects.HostRecord(
        connector=None,
        aliases=[mock_fqdn],
        comment=faker.sentence(),
        ipv4addrs=[
            objects.IPv4(
                ipv4addr=str(mock_v4),
                configure_for_dhcp=False,
                mac="00:00:00:00:00:00",
                ip=str(mock_v4),
                host=f"lo0.{mock_fqdn}",
            ),
        ],
        name=mock_fqdn,
    )
    mock_find_network_by_cidr.return_value = objects.NetworkV4(
        connector=None,
        comment=faker.sentence(),
        network=str(mock_v4_net),
        network_view="default",
        cidr=str(mock_v4_net),
    )

    #  Run workflow
    product_id = get_product_id_by_name(ProductName.ROUTER)
    initial_router_data = [{"product": product_id}, router_creation_input_form_data]
    result, process_stat, step_log = run_workflow("create_router", initial_router_data)

    result, step_log = assert_lso_interaction_success(result, process_stat, step_log)

    state = extract_state(result)
    subscription_id = state["subscription_id"]
    subscription = Router.from_subscription(subscription_id)

    assert_lso_interaction_failure(result, process_stat, step_log)

    assert subscription.status == "initial"
    assert subscription.description == f"Router {mock_fqdn}"

    assert mock_provision_router.call_count == 2
    assert mock_netbox_create_device.call_count == 0
    assert mock_find_host_by_fqdn.call_count == 0
    assert mock_find_network_by_cidr.call_count == 0
